import numpy as np
from zw_outliersdetec.__sax_via_window import *

'''
函数功能：计算欧式距离
'''
def euclidean(a, b):
    """Compute a Euclidean distance value."""
    return np.sqrt(np.sum((a-b)**2))

'''
类功能：实现HOT_SAX算法，找出异常时间序列
'''
class HOTSAX:
    #初始化，num_discords-预设输出异常的数量
    def __init__(self,num_discords=2):
        self.num_discords=num_discords

    '''
    功能：利用HOT-SAX算法找出异常时间序列的位置信息
    输入：series-数据集，win-size-自行设置的窗口大小（可理解为序列长度，默认为100），其他参数默认
    输出：根据设定的异常个数输出异常的开始位置以及对应的分数，表示从该位置开始的长度为win-size的序列被认为是异常
    运行示例：
        import numpy as np
        from HOTSAX import *
        #ECG数据，不带标签，只有一列值
        data = np.genfromtxt("ECG0606_1.csv", delimiter=',')
        
        hs=HOTSAX(2)
        discords,win_size =hs.find_discords_hotsax(data)
        print(discords,win_size)
    '''
    def hotsax(self,series, win_size=100, a_size=3,
                             paa_size=3, z_threshold=0.01):
        """HOT-SAX-driven discords discovery."""
        discords = list()

        globalRegistry = set()

        while (len(discords) < self.num_discords):

            bestDiscord =self.find_best_discord_hotsax(series, win_size, a_size,
                                                   paa_size, z_threshold,
                                                   globalRegistry)

            if -1 == bestDiscord[0]:
                break

            discords.append(bestDiscord)

            mark_start = bestDiscord[0] - win_size
            if 0 > mark_start:
                mark_start = 0

            mark_end = bestDiscord[0] + win_size
            '''if len(series) < mark_end:
                mark_end = len(series)'''

            for i in range(mark_start, mark_end):
                globalRegistry.add(i)

        return discords,win_size #返回设定异常个数的异常开始位置和窗口大小


    def find_best_discord_hotsax(self,series, win_size, a_size, paa_size,
                                 znorm_threshold, globalRegistry): # noqa: C901
        """Find the best discord with hotsax."""
        """[1.0] get the sax data first"""
        sax_none = sax_via_window(series, win_size, a_size, paa_size, "none", 0.01)

        """[2.0] build the 'magic' array"""
        magic_array = list()
        for k, v in sax_none.items():
            magic_array.append((k, len(v)))

        """[2.1] sort it desc by the key"""
        m_arr = sorted(magic_array, key=lambda tup: tup[1])

        """[3.0] define the key vars"""
        bestSoFarPosition = -1
        bestSoFarDistance = 0.

        distanceCalls = 0

        visit_array = np.zeros(len(series), dtype=np.int)

        """[4.0] and we are off iterating over the magic array entries"""
        for entry in m_arr:

            """[5.0] some moar of teh vars"""
            curr_word = entry[0]
            occurrences = sax_none[curr_word]

            """[6.0] jumping around by the same word occurrences makes it easier to
            nail down the possibly small distance value -- so we can be efficient
            and all that..."""
            for curr_pos in occurrences:

                if curr_pos in globalRegistry:
                    continue

                """[7.0] we don't want an overlapping subsequence"""
                mark_start = curr_pos - win_size
                mark_end = curr_pos + win_size
                visit_set = set(range(mark_start, mark_end))

                """[8.0] here is our subsequence in question"""
                cur_seq = znorm(series[curr_pos:(curr_pos + win_size)],
                                znorm_threshold)

                """[9.0] let's see what is NN distance"""
                nn_dist = np.inf
                do_random_search = 1

                """[10.0] ordered by occurrences search first"""
                for next_pos in occurrences:

                    """[11.0] skip bad pos"""
                    if next_pos in visit_set:
                        continue
                    else:
                        visit_set.add(next_pos)

                    """[12.0] distance we compute"""
                    dist = euclidean(cur_seq, znorm(series[next_pos:(
                                     next_pos+win_size)], znorm_threshold))
                    distanceCalls += 1

                    """[13.0] keep the books up-to-date"""
                    if dist < nn_dist:
                        nn_dist = dist
                    if dist < bestSoFarDistance:
                        do_random_search = 0
                        break

                """[13.0] if not broken above,
                we shall proceed with random search"""
                if do_random_search:
                    """[14.0] build that random visit order array"""
                    curr_idx = 0
                    for i in range(0, (len(series) - win_size)):
                        if not(i in visit_set):
                            visit_array[curr_idx] = i
                            curr_idx += 1
                    it_order = np.random.permutation(visit_array[0:curr_idx])
                    curr_idx -= 1

                    """[15.0] and go random"""
                    while curr_idx >= 0:
                        rand_pos = it_order[curr_idx]
                        curr_idx -= 1

                        dist = euclidean(cur_seq, znorm(series[rand_pos:(
                                         rand_pos + win_size)], znorm_threshold))
                        distanceCalls += 1

                        """[16.0] keep the books up-to-date again"""
                        if dist < nn_dist:
                            nn_dist = dist
                        if dist < bestSoFarDistance:
                            nn_dist = dist
                            break

                """[17.0] and BIGGER books"""
                if (nn_dist > bestSoFarDistance) and (nn_dist < np.inf):
                    bestSoFarDistance = nn_dist
                    bestSoFarPosition = curr_pos

        return (bestSoFarPosition, bestSoFarDistance)



import numpy
from sklearn.preprocessing import scale

class PAPR:
    splits = dict() #创建一个新的字典
    splits[2] = [0, float('inf')]   #float（‘inf’）正无穷
    splits[3] = [-0.43, 0.43, float('inf')]
    splits[4] = [-0.67, 0, 0.67, float('inf')]
    splits[5] = [-0.84, -0.25, 0.25, 0.84, float('inf')]
    splits[6] = [-0.97, -0.43, 0, 0.43, 0.97, float('inf')]
    splits[7] = [-1.07, -0.57, -0.18, 0.18, 0.57, 1.07, float('inf')]
    splits[8] = [-1.15, -0.67, -0.32, 0, 0.32, 0.67, 1.15, float('inf')]
    splits[9] = [-1.22, -0.76, -0.43, -0.14, 0.14, 0.43, 0.76, 1.22, float('inf')]
    splits[10] = [-1.28, -0.84, -0.52, -0.25, 0, 0.25, 0.52, 0.84, 1.28, float('inf')]

    def __init__(self,m):
        self.m=m

    '''
    功能：利用PAPR_RW算法进行异常分数的计算
    输入：data，从CSV文件读取的一维数据
    输出：data_matrix，二维矩阵，每行为一个子序列
    运行示例：
        from PAPR import *
        import numpy as np
        import matplotlib.pyplot as plt
        
        m=250  #设置子序列长度,每个数据集不一样
        P=PAPR(m)
        #从CSV文件读取数据，不带标签
        data = np.genfromtxt("chfdb_1.csv", delimiter=',')
        
        scores,scores_index=P.PAPR(data)
        
        #可视化标注异常
        t=scores_index[0]   #得分最低的序列的下标，该序列被认为是异常
        x0=range(0,data.__len__())
        plt.plot(x0,data)
        x1=range(t*m,t*m+m-1)
        plt.plot(x1,data[t*m:t*m+m-1],color="red")
        plt.show()
    
    '''
    def papr(self,data):
        m=self.m  #划分的序列长度
        # 将数据划分成多个子序列
        data_matrix = self.split_data(data)
        # 对子序列分别进行归一化处理
        data_matrix = scale(data_matrix, axis=1)
        # 计算子序列的高斯径向基函数的宽度参数
        widths = self.find_best_w(data_matrix)
        matrix = self.cal_matrix(data_matrix, 6)
        sim_matrix = self.cal_similarity(matrix=matrix, wc=0.3, wd=0.4, wr=0.3, length=widths.__len__(), widths=widths)
        scores = self.random_walk(sim_matrix, error=0.05)

        scores_index=numpy.argsort(scores) #异常分数从低到高排序

        return scores,scores_index


    '''
    输入：data，一维数据
    功能：将读取到的一维数据按照传入的子序列长度进行划分
    输出：data_matrix，二维矩阵，每行为一个子序列
    '''
    def split_data(self,data):
        index=self.m
        length=data.__len__()
        data_matix=list()
        sequence=list()
        i=0
        while i<length:
            sequence=data[i:i+index]
           # print(sequence)
            i=i+index
            data_matix.append(sequence)
        return data_matix

    '''
    值空间划分以及PAPR指标的计算
    '''
    def cal_matrix(self,data, k):
        points = self.splits[k]
        new_data = list()
        for item in data:
            tmp_points = list()
            for i in range(k):
                tmp_points.append(list())
            for p in item:
                for w in range(k):
                    if p < points[w]:
                        tmp_points[w].append(p)
                        break
            tmp_matrix = numpy.zeros((k, 3))    #生成一个K行3列的全0矩阵，用于记录PAPR方法得出的Mi = [di, ci, ri]
            for w in range(k):
                tmp_matrix[w, 0] = len(tmp_points[w])   #记录子值空间的点个数
                if tmp_matrix[w, 0] != 0:
                    tmp_matrix[w, 1] = numpy.mean(tmp_points[w])     #记录子值空间的点均值
                    tmp_matrix[w, 2] = numpy.var(tmp_points[w])     #记录子值空间的方差
            new_data.append(tmp_matrix)

        return numpy.array(new_data)

    '''
    计算相似度矩阵
    #length是子序列的数量，width是计算Scij和Srij使所用到的δ
    '''
    def cal_similarity(self,matrix, length, wd, wc, wr, widths):
        index = range(length)
        sim_matrix = numpy.zeros((length, length))   #生成一个length行length列的全0矩阵
        for r in index:
            for c in index:
                sd = self.cal_d_sim(matrix[r, :, 0], matrix[c, :, 0])
                sc = self.cal_rc_sim(matrix[r, :, 1], matrix[c, :, 1], widths[r])
                sr = self.cal_rc_sim(matrix[r, :, 2], matrix[c, :, 2], widths[r])
                sim_matrix[r, c] = wd*sd + wc*sc + wr*sr
        return sim_matrix

    '''
    函数功能：计算记录两点数量的向量di和dj的相似度Sdij
    '''
    def cal_d_sim(self,one, two):
        #m是子序列one的总长度，m=∑（k=1..q）dik
        m = numpy.sum(one)
        #length是记录子序列特征的Mi=[di,ci,ri]的长度,即子值空间的划分数目
        length = len(one)
        s = 0
        for l in range(length):
            s += min(one[l], two[l])
        return 1.0 * s / m

    '''
    函数功能：计算Scij和Srij，两个计算公式相同
    w即δ，高斯径向基函数的半径，通过信息熵的方法可以计算出每个数据集的该值
    '''
    def cal_rc_sim(self,one, two, w=0.005):
        return numpy.exp(-1.0 * numpy.linalg.norm(one - two, ord=2) / numpy.power(w, 2))

    '''
    RW模型,最终会得到一个概率分布矩阵，即异常得分
    '''
    def random_walk(self,sim_matrix, error=0.1):
        rows, cols = sim_matrix.shape
        s_matrix = numpy.zeros((rows, cols))
        for r in range(rows):
            totSim = 0.0
            for c in range(cols):
                totSim += sim_matrix[r, c]
            for c in range(cols):
                s_matrix[r, c] = 1.0*sim_matrix[r, c] / totSim

        damping_factor = 0.1
        ct = numpy.array([1.0/rows]*rows)
        recursive_err = error+1
        times = 0
        while recursive_err > error and times < 100:
            ct1 = damping_factor/rows + numpy.dot(s_matrix.T, ct)
            recursive_err = numpy.linalg.norm(ct-ct1, ord=1)
            times += 1
            ct = ct1[:]
        return ct

    '''
    函数功能：计算数据集的δ，高斯径向基函数的半径，通过信息熵的方法计算
    '''
    def find_best_w(self,data_matrix):

        alist, blist = numpy.zeros(data_matrix.__len__()), numpy.zeros(data_matrix.__len__())
        r_index = range(data_matrix.__len__())
        gama = (5**0.5-1)/2
        coe = (2**0.5)/3
        for i in r_index:
            min_dist, max_dist = float('inf'), -float('inf')
            for j in r_index:
                if i == j:
                    continue
                dist = numpy.linalg.norm(data_matrix[i]-data_matrix[j], ord=2)    #求二范数
                min_dist = min(dist, min_dist)
                max_dist = max(dist, max_dist)
            alist[i], blist[i] = coe*min_dist, coe*max_dist
        left, right = cal_sig(alist, blist, gama)
        ent_left = cal_entropy(left)
        ent_right = cal_entropy(right)
        epison = 1
        times = 0
        while numpy.linalg.norm(alist-blist) < 1 and times < 20:
            if ent_left < ent_right:
                blist, right = right.copy(), left.copy()
                ent_right = ent_left
                left = alist + (1-gama)*(blist-alist)
                ent_left = cal_entropy(left)
            else:
                alist, left = left.copy(), right.copy()
                ent_left = ent_right
                right = alist + gama*(blist-alist)
                ent_right = cal_entropy(right)

            times += 1

        if ent_left < ent_right:
            return left
        else:
            return right


def cal_sig(alist, blist, gama):
    length = len(alist)
    index = range(length)
    left, right = numpy.zeros(length), numpy.zeros(length)
    for i in index:
        left[i] = alist[i] + (1-gama)*(blist[i]-alist[i])
        right[i] = alist[i] + gama*(blist[i]-alist[i])
    return left, right

'''
计算信信息熵
'''
def cal_entropy(list):
    total = sum(list)
    list /= total
    log_list = numpy.log(list)
    return -numpy.dot(list, log_list)




import numpy as np
'''
函数功能：PAA算法
'''
def paa(series, paa_segments):
    """PAA implementation."""
    series_len = len(series)

    # check for the trivial case
    if (series_len == paa_segments):
        return np.copy(series)
    else:
        res = np.zeros(paa_segments)
        # check when we are even
        if (series_len % paa_segments == 0):
            inc = series_len // paa_segments
            for i in range(0, series_len):
                idx = i // inc
                np.add.at(res, idx, series[i])
                # res[idx] = res[idx] + series[i]
            return res / inc
        # and process when we are odd
        else:
            for i in range(0, paa_segments * series_len):
                idx = i // series_len
                pos = i // paa_segments
                np.add.at(res, idx, series[pos])
                # res[idx] = res[idx] + series[pos]
            return res / series_len


import numpy as np
import math
from operator import add
from zw_outliersdetec import __dimension_reduction as dim_red

class FastVOA:
    #t是随机超平面投影时的参数
    def __init__(self,t):
        self.t=t

    #Algorithm 3 FirstMomentEstimator(L; t; n)
    def __first_moment_estimator(self,projected, t, n):
        f1 = [0] * n
        for i in range(0, t):
            cl = [0] * n
            cr = [0] * n
            li = projected[i]
            for j in range(0, n):
                idx = li[j][0]
                cl[idx] = j     #原：cl[idx] = j - 1
                cr[idx] = n - 1 - cl[idx]
            for j in range(0, n):
                f1[j] += cl[j] * cr[j]
        return list(map(lambda x: x * ((2 * math.pi) / (t * (n - 1) * (n - 2))), f1))

    #Algorithm 4 FrobeniusNorm(L; t; n)
    def __frobenius_norm(self,projected, t, n):
        f2 = [0] * n
        sl = np.random.choice([-1, 1], size=(n,), p=None)
        sr = np.random.choice([-1, 1], size=(n,), p=None)
        for i in range(0, t):
            amsl = [0] * n
            amsr = [0] * n
            li = projected[i]
            for j in range(1, n):
                idx1 = li[j][0]
                idx2 = li[j - 1][0]
                amsl[idx1] = amsl[idx2] + sl[idx2]
            for j in range(n - 2, -1, -1):
                idx1 = li[j][0]
                idx2 = li[j + 1][0]
                amsr[idx1] = amsr[idx2] + sr[idx2]
            for j in range(0, n):
                f2[j] += amsl[j] * amsr[j]
        return f2

    #Algorithm 1 FastVOA(S; t; s1; s2)
    '''
    功能：计算每个数据对象的角度值
    输入：train-不带标签的数据，n-数据对象的个数，t-随机超平面投影参数，s1,s2
    输出：每个数据对象的角度值分数scores以及按分数排序后的下标scores_index
    调用示例：
    
        import pandas as pd
        from FastVOA import *
        
        #读取数据,不需要标签
        data=pd.read_csv('isolet.csv',sep=',')
        ytrain = data.iloc[:, -1]
        train=data.drop('Label',axis=1)
        #对数据进行随机超平面投影
        DIMENSION = 600
        t = DIMENSION
        n = train.shape[0]
        print(n)
        #调用FatsVOA进行角度值计算
        fv=FastVOA(t)
        scores,scores_index=fv.fastvoa(train, n, t, 1, 1)
    '''
    def fastvoa(self,train, n, t, s1, s2):
        projected = dim_red.random_projection(train, t)
        f1 = self.__first_moment_estimator(projected, t, n)
        y = []
        for i in range(0, s2):
            s = [0] * n
            for j in range(0, s1):
                result = list(map(lambda x: x ** 2, self.__frobenius_norm(projected, t, n)))
                s = list(map(add, s, result))
            s = list(map(lambda x: x / s1, s))  #s的长度为n（由于result的长度为n）
            y.append(s)     #yi的长度为n，y的长度为s2，y有s2行，n列
        y = list(map(list, zip(*y))) #拆分y
        f2 = []
        for i in range(0, n):
            f2.append(np.average(y[i])) #求均值
        var = [0] * n
        for i in range(0, n):
            f2[i] = (4 * (math.pi ** 2) / (t * (t - 1) * (n - 1) * (n - 2))) * f2[i] - (2 * math.pi * f1[i]) / (t - 1)
            var[i] = f2[i] - (f1[i] ** 2)
        # 排序
        scores=var
        scores_index = np.argsort(scores)  # 按角度值从低到高排序
        return scores,scores_index[::-1]  #返回异常得分及下标



import numpy
from interval import *
from math import *
class IntervalSets:
    #m为子序列长度，可以自行设置，默认为100
    def __init__(self,m=100):
        self.m=m
    '''
    功能：利用IntervalSets方法计算出每个序列的异常结构分数
    输入：data-从CSV文件读取的数据信息，不含标签，一维
    输出：每个序列的异常结构分数以及排序后的下标
    运行示例：
        import numpy as np
        import matplotlib.pyplot as plt
        from IntervalSets import *
        
        m=400  #设置子序列长度，ECG为400，chfdb为250时效果最佳
        #创建对象
        IS=IntervalSets(m)
        #从CSV文件读取数据
        data= np.genfromtxt('ECG108_2.csv',delimiter=',')
        #调用IntervalSets方法来查找异常
        SAS,SAS_index=IS.intervalsets(data)
        print(SAS)  #输出异常结构分数SAS
        print("最可能的异常序列标号为：",SAS_index[len(SAS_index)-1])
        
        #可视化标注异常
        x0=range(0,data.__len__())
        plt.plot(x0,data)
        t=SAS_index[len(SAS_index)-1]
        x1=range(t*m,t*m+m-1)
        plt.plot(x1,data[t*m:t*m+m-1],color="red")
        plt.text(100, 1,"m=400", size = 15, alpha = 0.8)
        plt.show()
        
    '''
    def intervalsets(self,data):
        data_matrix=self.split_data(data)   #对数据进行划分，形成多个子序列
        n=len(data_matrix) #子序列的个数
        m=self.m    #子序列长度
        Sp=self.cal_Sp(data_matrix)
        SA=self.cal_SA(data_matrix)
        #计算每个子序列的异常结构分数
        S=list()    #二维
        SAS=list()  #存储分数，一维
        for i in range(n):
            S_i=0
            for j in range(n):
                S_i=S_i+(0.5*SA[i][j]+0.5*Sp[i][j])
            S.append(S_i)
        for i in range(n):
            SAS_i =0
            for j in range(n):
                SAS_i=SAS_i+(S[i]-S[j])**2
            SAS.append(SAS_i/n)
        #排序
        SAS_index = numpy.argsort(SAS)

        return SAS,SAS_index

    '''
       输入：data_matrix，二维数据
       功能：类的私有方法，计算子序列之间的概率相似性Sp,boundary为区间边界参数
       输出：Sp，二维矩阵，每行为一个子序列与其他子序列的Spij值
    '''
    def cal_Sp(self,data_matrix,boundary=0.2):
        Sp=list()
        Amin = list()  # 每一行存储一个子序列的最小值
        Amax = list()  # 每一行存储一个子序列的最大值
        Pmin = list()  # 每一行存储一个子序列的概率最小值
        Pmax = list()  # 每一行存储一个子序列的概率最大值
        n = len(data_matrix)  # 子序列数量

        widths=self.find_best_w(data_matrix)
        #求子序列的取值区间
        for i in range(n):
            Amin.append(min(data_matrix[i]))  # 子序列最小值
            Amax.append(max(data_matrix[i]))  # 子序列最大值
        #求点分布在边界区间的概率
        for i in range(n):
            count_min=0
            count_max = 0
            for item in data_matrix[i]:
                if item>=Amin[i] and item<=Amin[i]+boundary*(Amax[i]-Amin[i]):
                    count_min=count_min+1
                if item>=Amax[i]-boundary*(Amax[i]-Amin[i]) and item<=Amax[i]:
                    count_max = count_max + 1

            Pmin.append(count_min/self.m)
            Pmax.append(count_max/self.m)

        #利用边界的点分布概率计算Sp
        for i in range(n):
            Sp_i=list()
            for j in range(n):
                if i==j:
                    Sp_i.append(1)
                else:
                    p=exp(-((Pmin[i]-Pmin[j])**2+(Pmax[i]-Pmax[j])**2)/widths[i]**2)
                    #p=numpy.exp(-1.0 * numpy.linalg.norm(one - two, ord=2) / numpy.power(w, 2))
                    Sp_i.append(p)
            Sp.append(Sp_i)
        return Sp



    '''
       输入：data_matrix，二维数据
       功能：类的私有方法，计算子序列之间的概率相似性SA
       输出：SA，二维矩阵，每行为一个子序列与其他子序列的SAij值
    '''

    def cal_SA(self, data_matrix):
        Amin=list()    #每一行存储一个子序列的最小值
        Amax = list()  # 每一行存储一个子序列的最大值
        SA=list()   #存储区相似度
        n=len(data_matrix) #子序列数量

        for i in range(n):
            Amin.append(min(data_matrix[i]))  # 子序列最小值
            Amax.append(max(data_matrix[i]))  # 子序列最大值

        for i in range(n):
            SA_i=list()
            A_i=Interval(Amin[i],Amax[i])
            #print(A_i)
            for j in range(n):
                A_j=Interval(Amin[j],Amax[j])
                if A_i.overlaps(A_j)==False:    #情况1，没有交集
                    SA_i.append(0)
                else:   #情况2，有交集
                    A_ij=A_i.join(A_j)  #合并
                    a=((A_i.upper_bound-A_i.lower_bound)+(A_j.upper_bound-A_j.lower_bound)-(A_ij.upper_bound-A_ij.lower_bound))/(A_ij.upper_bound-A_ij.lower_bound)
                    SA_i.append(a)
            SA.append(SA_i)

        return SA

    '''
       输入：data，一维数据
       功能：将读取到的一维数据按照传入的子序列长度进行划分
       输出：data_matrix，二维矩阵，每行为一个子序列
    '''
    def split_data(self, data):
        index = self.m
        length = data.__len__()
        data_matix = list()
        sequence = list()
        i = 0
        while i < length:
            sequence = data[i:i + index]
            # print(sequence)
            i = i + index
            data_matix.append(sequence)
        return data_matix

    '''
    函数功能：计算数据集的δ，高斯径向基函数的半径，通过信息熵的方法计算
    返回一个与子序列长度相等的数组
    '''
    def find_best_w(self, data_matrix):

        alist, blist = numpy.zeros(data_matrix.__len__()), numpy.zeros(data_matrix.__len__())
        r_index = range(data_matrix.__len__())
        gama = (5 ** 0.5 - 1) / 2
        coe = (2 ** 0.5) / 3
        for i in r_index:
            min_dist, max_dist = float('inf'), -float('inf')
            for j in r_index:
                if i == j:
                    continue
                dist = numpy.linalg.norm(data_matrix[i] - data_matrix[j], ord=2)  # 求二范数
                min_dist = min(dist, min_dist)
                max_dist = max(dist, max_dist)
            alist[i], blist[i] = coe * min_dist, coe * max_dist
        left, right = cal_sig(alist, blist, gama)
        ent_left = cal_entropy(left)
        ent_right = cal_entropy(right)
        epison = 1
        times = 0
        while numpy.linalg.norm(alist - blist) < 1 and times < 20:
            if ent_left < ent_right:
                blist, right = right.copy(), left.copy()
                ent_right = ent_left
                left = alist + (1 - gama) * (blist - alist)
                ent_left = cal_entropy(left)
            else:
                alist, left = left.copy(), right.copy()
                ent_left = ent_right
                right = alist + gama * (blist - alist)
                ent_right = cal_entropy(right)

            times += 1

        if ent_left < ent_right:
            return left
        else:
            return right


def cal_sig(alist, blist, gama):
    length = len(alist)
    index = range(length)
    left, right = numpy.zeros(length), numpy.zeros(length)
    for i in index:
        left[i] = alist[i] + (1 - gama) * (blist[i] - alist[i])
        right[i] = alist[i] + gama * (blist[i] - alist[i])
    return left, right

'''
计算信信息熵
'''
def cal_entropy(list):
    total = sum(list)
    list /= total
    log_list = numpy.log(list)
    return -numpy.dot(list, log_list)





from setuptools import setup,find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="zw_outliersdetec",
    version="0.0.1",    #第一版
    author="Zuo Wei",
    author_email="zw@my.swjtu.edu.cn",
    description="A small outliers-detection package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vidiewei",
    packages=find_packages(),
    #include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


import numpy as np
from sklearn.neighbors import NearestNeighbors
from math import *

'''
说明：RDOS算法，根据K近邻集合、逆近邻集合以及共享近邻集合进行对象的核密度估计，从而得出异常分数
传入参数：k-近邻的个数，h-高斯核的宽度参数
'''
class RDOS:
    #初始化，传入参数并设置默认值
    def __init__(self,n_outliers=1,n_neighbors=5,h=2):
        self.n_outliers=n_outliers
        self.n_neighbors=n_neighbors
        self.h=h

    '''
    RDOS
    输入：data-数据集，n-返回的异常个数(默认为1)，n_neighbors-近邻个数，h-高斯核函数的宽度参数
    输出：返回异常的分数以及按异常分数排序好的下标，同时会根据预设的异常个数输出其下标及得分
    运行示例：
        import pandas as pd
        from RDOS import *
        
        #如果是带有Lable的数据集则需要先去除Label列
        data=pd.read_csv("hbk.csv",sep=',')
        #data=data.drop('Label', axis=1)
        data=np.array(data)
        print(data.shape)
        #调用RDOS算法，传入预设的异常个数
        rdos=RDOS(n_outliers=10)
        RDOS_index,RDOS_score=rdos.rdos(data)
    '''
    def rdos(self,data):
        n_outliers = self.n_outliers
        n_neighbors=self.n_neighbors
        h=self.h
        n=data.shape[0] #n是数据集的样例数量
        d=data.shape[1] #d是数据集的维度，即属性个数

        #规范输入的参数
        if n_neighbors>=n or n_neighbors<1:
            print('n_neighbors input must be less than number of observations and greater than 0')
            exit()

        outliers=list()
        #存储每个数据对象的近邻下标
        Sknn= list()
        Srnn= list()
        Ssnn = list()
        S= list()
        P= list()
        #计算Sknn
        for X in data:
            Sknn_temp = self.KNN(data, [X], return_distance=False)
            Sknn_temp = np.squeeze(Sknn_temp)
            Sknn.append(Sknn_temp[1:])
            S.append(list(Sknn_temp[1:]))  # X的所有近邻集合
        #计算Srnn
        for i in range(n):
            Srnn_temp = list()  # 记录每个数据对象的rnn
            for item in Sknn[i]:
                item_neighbors = Sknn[item]
                # 如果X的近邻的k近邻集合中也包含X，说明该近邻是X的逆近邻
                if i in item_neighbors:
                    Srnn_temp.append(item)
            Srnn.append(Srnn_temp)
            S[i].extend(Srnn_temp)  # X的所有近邻集合
        #计算Ssnn
        for i in range(n):
            Ssnn_temp = list()
            for j in Sknn[i]:
                kneighbor_rnn = Srnn[j]  # k近邻的逆近邻集合
                Ssnn_temp.extend(kneighbor_rnn)
            Ssnn_temp = list(set(Ssnn_temp))  # 去重
            if i in Ssnn_temp:
                Ssnn_temp.remove(i)  # 删除X本身下标
            Ssnn.append(Ssnn_temp)  # X的共享近邻集合
            S[i].extend(Ssnn_temp)  # X的所有近邻集合
            S[i] = list(set(S[i]))  # 去重

            P.append(self.getKelnelDensity(data, i, S[i]))#计算论文中的P值

        '''
        #计算每个数据对象的近邻集合
        for i in range(n):
            Sknn_temp=self.KNN(data,[data[i]],return_distance=False)
            Sknn_temp = np.squeeze(Sknn_temp)
            print("Sknn:",Sknn_temp[1:])
            Sknn.append(Sknn_temp[1:]) #需要除去其本身,例：[[11 29  7 26 24]]→[29  7 26 24]
            Srnn.append(self.RNN(data,[data[i]],return_distance=False)) #例：[29 24]
            Ssnn_temp=list()
            for j in Sknn[i]:
                kneighbor_rnn = self.RNN(data, [data[j]], return_distance=False) #k近邻的逆近邻集合
                Ssnn_temp.extend(kneighbor_rnn)
            Ssnn_temp = list(set(Ssnn_temp))  # 去重
            if i in Ssnn_temp:
                Ssnn_temp.remove(i)  # 删除X本身下标
            Ssnn.append(Ssnn_temp) #X的共享近邻集合
            S.append(list(set(Ssnn_temp)))  #X的所有近邻集合
            '''

            #print("S:",S[i]) #打印


        #计算异常得分RDOS
        RDOS_score=list()

        for i in range(n):
            S_RDOS=0
            for j in S[i]:  #计算近邻集合的RDOS总分数
                S_RDOS=S_RDOS+P[j]

            RDOS_score.append(S_RDOS/(len(S[i])*P[i]))

        RDOS_index= np.argsort(RDOS_score) #对异常分数进行排序，从低到高,返回的是数组的索引


        return RDOS_score,RDOS_index[::-1]  #返回异常的得分及其下标（下标由得分从高到低排序）



    '''
    找出数据集X中每个对象的的k近邻并返回序号(当k>1时，会包括其本身)
    X可以是一个点或者一组数据,data是所有数据
    return_distance=True时会同时返回距离
    '''
    def KNN(self,data,X,return_distance=False):
        neigh = NearestNeighbors(self.n_neighbors)
        neigh.fit(data)
        return neigh.kneighbors(X, return_distance=return_distance)

    '''
    找出X的k逆近邻集合并返回序号
    X是一个数据对象,data是所有数据
    return_distance=True时会同时返回距离
    
    def RNN(self, data, X, return_distance=False):
        neigh = NearestNeighbors(self.n_neighbors)
        neigh.fit(data)
        X_neighbors=neigh.kneighbors(X, return_distance=return_distance)
        X_Srnn=list()   #存储逆近邻的下标集合
        # 遍历X的近邻集合寻找其逆近邻集合，item为近邻的序号
        index = X_neighbors[0, 1:]
        X_index = X_neighbors[0, 0] #X的下标
        #近邻的下标
        for item in index:
            item_neighbors = neigh.kneighbors([data[item]], return_distance=False)  #寻找近邻的k近邻集合
            # 如果X的近邻的k近邻集合中也包含X，说明该近邻是X的逆近邻
            if X_index in item_neighbors:
                X_Srnn.append(item)
        return np.array(X_Srnn)
    '''

    '''
    计算核密度
    输入：data-数据集，X_index-数据对象的下标，S近邻集合
    输出：论文中的P
    '''
    def getKelnelDensity(self,data,X_index,S):
        h=self.h    #高斯核函数参数
        d=data.shape[1] #数据的属性个数
        S_X=list(S)
        S_X.append(X_index)
        X_guassian =0
        for i in S_X:
            X_guassian+=(1/((2*pi)**(d/2)))*exp(-(np.linalg.norm(data[i]-data[X_index]))/(2*h**2))

        S_len=S.__len__()
        P=1/(S_len+1)*(1/h**d)*X_guassian
        return P












import random
import numpy as np
from sklearn.decomposition import PCA
import pandas as pd
from sklearn.decomposition import TruncatedSVD

#Algorithm 2 RandomProjection(S; t)
def random_projection(S, t):
    """
    Projects data space to random vector space
    :param S: Train set
    :param t: Amount of final dimension
    :return: A list L = L1L2L3...Lt where Li is a list of points
            ordered by their dot product with ri
    """
    l = []
    for i in range(0, t):
        #ri = []
        #for j in range(0, S.shape[1]):
            #ri.append(random.randint(0, 1))
        ri=list(np.random.normal(0,1,S.shape[1]))
        l.append([])
        for index, record in S.iterrows():
            dotted = np.dot(record, ri)
            l[i].append((index, dotted))
        l[i] = sorted(l[i], key=lambda x: x[1])
    return l


def prepare_projected_data(projected, t):
    result = list()
    for i in range(0, t):
        l = sorted(projected[i], key=lambda x: x[0])
        l = list(map(lambda x: x[1], l))
        result.append(l)
    result = list(map(list, zip(*result)))
    return result


def pca(S, t, is_product):
    pca = PCA(n_components=t)
    X = np.array(S)
    pca.fit(X)
    if not is_product:
        print("Accumulative Variance Ratio : ", pca.explained_variance_ratio_.cumsum())
    return pd.DataFrame(pca.transform(X))


def SVD(S, t, is_product):
    svd = TruncatedSVD(n_components=t)
    X = np.array(S)
    svd.fit(X)
    if not is_product:
        print("Accumulative Variance Ratio : ", svd.explained_variance_ratio_.cumsum())
    return pd.DataFrame(svd.transform(X))